#include "StdAfx.h"
#include "Window.h"

extern HINSTANCE g_hInstance;

Window::Window() throw() : Handle(NULL),MenuHandle(NULL),TextFont(NULL),
    PositionX(0),PositionY(0),Width(0),Height(0),TextColor(RGB(0,0,0)),
    BackColor(RGB(255,255,255)),Text(),Controls()
{
    TextFont = ::CreateFontW(12,5,0,0,400,FALSE,FALSE,FALSE,ANSI_CHARSET,
        OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY,DEFAULT_PITCH | FF_DONTCARE,
        L"MS Sans Serif");
}

Window::~Window() throw()
{
    if (MenuHandle != NULL)
    {
        ::DestroyMenu(MenuHandle);
    }
    
    if (TextFont != NULL)
    {
        ::DeleteObject(TextFont);
    }
}

bool Window::Create(const String & strName,const String & strTitle,
    unsigned long dwStyle,unsigned long dwExtendedStyle,long x,long y,
    long nWidth,long nHeight,unsigned long dwID,Window * pParent) throw()
{
    if (Handle == NULL)
    {
        MDICREATESTRUCTW mdiCreate;
        mdiCreate.szClass = strName.c_str();
        mdiCreate.szTitle = strTitle.c_str();
        mdiCreate.hOwner = ::g_hInstance;
        mdiCreate.x = x;
        mdiCreate.y = y;
        mdiCreate.cx = nWidth;
        mdiCreate.cy = nHeight;
        mdiCreate.style = dwStyle;
        mdiCreate.lParam = reinterpret_cast<LPARAM>(this);
        
        HWND hParentWindow = pParent != NULL ? pParent->Handle : NULL;
        
        Handle = ::CreateWindowExW(dwExtendedStyle,strName.c_str(),strTitle.c_str(),
            dwStyle,x,y,nWidth,nHeight,hParentWindow,MenuHandle,NULL,&mdiCreate);
        
        if (Handle != NULL)
        {
            if (pParent != NULL)
            {
                pParent->Controls[dwID] = this;
            }
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
}

bool Window::Show(int nShow) throw()
{
    if (Handle != NULL)
    {
        return ::ShowWindow(Handle,nShow) != 0;
    }
    else
    {
        return false;
    }
}

bool Window::Update() throw()
{
    if (Handle != NULL)
    {
        return ::UpdateWindow(Handle) != 0;
    }
    else
    {
        return false;
    }
}

bool Window::Invalidate(bool bEraseNow) throw()
{
    if (Handle != NULL)
    {
        const BOOL bErase = bEraseNow == true ? TRUE : FALSE;
        return ::InvalidateRect(Handle,NULL,bErase) != 0;
    }
    else
    {
        return false;
    }
}

bool Window::Redraw(bool bRedrawChildren) throw()
{
    if (Handle != NULL)
    {
        unsigned long dwRedrawValue = RDW_ERASE | RDW_INVALIDATE | RDW_FRAME | RDW_INTERNALPAINT | RDW_UPDATENOW;
        if (bRedrawChildren == true)
        {
            dwRedrawValue |= RDW_ALLCHILDREN;
        }
        return ::RedrawWindow(Handle,NULL,NULL,dwRedrawValue) != 0;
    }
    else
    {
        return false;
    }
}

Window * Window::GetParent() const throw()
{
    if (Handle != NULL)
    {
        HWND hParent = ::GetParent(Handle);
        
        if (hParent != NULL)
        {
            return reinterpret_cast<Window *>(::GetPropW(Handle,L"WindowObject"));
        }
    }
    else
    {
        return NULL;
    }
}

HWND Window::GetHwnd() const throw()
{
    return Handle;
}

void Window::SetTextColor(COLORREF clrText) throw()
{
    TextColor = clrText;
}

COLORREF Window::GetTextColor() const throw()
{
    return TextColor;
}

void Window::SetBackColor(COLORREF clrBack) throw()
{
    BackColor = clrBack;
}

COLORREF Window::GetBackColor() const throw()
{
    return BackColor;
}

Window::String Window::GetText() const throw()
{
    return Text;
}

void Window::SetText(const String & str) throw()
{
    if (Handle != NULL)
    {
        ::SendMessageW(Handle,WM_SETTEXT,0,reinterpret_cast<LPARAM>(str.c_str()));
    }
    else
    {
        Text = str;
    }
}

Window * Window::GetChildWindow(unsigned long dwID) const throw()
{
    Window * pRetValue = NULL;
    try
    {
        ControlCollection & controlMap = const_cast<ControlCollection &>(Controls);
        pRetValue = controlMap[dwID];
    }
    catch(...)
    {
        return NULL;
    }
    return pRetValue;
}

bool Window::RegisterWindowClass(const String & strName,HCURSOR hCursor,HICON hIcon,
    HICON hIconSm) throw()
{
    WNDCLASSEXW wndClass;
    
    wndClass.cbSize = sizeof(WNDCLASSEXW);
    wndClass.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS | CS_GLOBALCLASS;
    wndClass.hInstance = ::g_hInstance;
    wndClass.lpfnWndProc = reinterpret_cast<WNDPROC>(Window::HandleWindowProcess);
    wndClass.cbWndExtra = 0;
    wndClass.cbClsExtra = 0;
    wndClass.lpszMenuName = NULL;
    wndClass.hbrBackground = NULL;
    
    wndClass.lpszClassName = strName.c_str();
    wndClass.hCursor = hCursor;
    wndClass.hIcon = hIcon;
    wndClass.hIconSm = hIconSm;
    
    return ::RegisterClassExW(&wndClass) != 0;
}

bool Window::UnregisterWindowClass(const String & strName) throw()
{
    return ::UnregisterClassW(strName.c_str(),::g_hInstance);
}

INT_PTR Window::WindowProcess(UINT message,WPARAM wParam,LPARAM lParam) throw()
{
    if (message == WM_NCCREATE)
    {
        return 1;
    }
    else if (message == WM_CREATE)
    {
        return 0;
    }
    else if (message == WM_SIZE)
    {
        if (wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED)
        {
            Width = LOWORD(lParam);
            Height = HIWORD(lParam);
            return 0;
        }
    }
    else if (message == WM_MOVE)
    {
        PositionX = LOWORD(lParam);
        PositionY = HIWORD(lParam);
        return 0;
    }
    else if (message == WM_SETTEXT)
    {
        const wchar_t * pString = reinterpret_cast<const wchar_t *>(lParam);
        Text = pString;
        return 1;
    }
    else if (message == WM_GETTEXT)
    {
        const size_t nBufferMax = wParam;
        wchar_t * pBuffer = reinterpret_cast<wchar_t *>(lParam);
        const size_t nTextLength = Text.length();
        const size_t nCopyLength = nTextLength < nBufferMax ? nTextLength : nBufferMax - 1;
        for (size_t x = 0; x < nCopyLength; ++x)
        {
            pBuffer[x] = Text[x];
        }
        pBuffer[nCopyLength] = 0;
        return nCopyLength;
    }
    else if (message == WM_GETTEXTLENGTH)
    {
        return Text.length();
    }
    else if (message == WM_SETFONT)
    {
        HFONT hFont = reinterpret_cast<HFONT>(wParam);
        if (hFont != NULL)
        {
            ::DeleteObject(TextFont);
            TextFont = hFont;
        }
        
        if (lParam == TRUE)
        {
            ::RedrawWindow(Handle,NULL,NULL,RDW_ERASE | RDW_INVALIDATE | RDW_FRAME | RDW_INTERNALPAINT | RDW_UPDATENOW | RDW_ALLCHILDREN);
        }
        
        return 0;
    }
    else if (message == WM_GETFONT)
    {
        return reinterpret_cast<INT_PTR>(TextFont);
    }
    else if (message == WM_CLOSE)
    {
        ::DestroyWindow(Handle);
    }
    else if (message == WM_NCDESTROY)
    {
        delete this;
    }
    else
    {
        return ::DefWindowProc(Handle,message,wParam,lParam);
    }
}

void Window::EraseBackground(HDC hDC) throw()
{
    HBRUSH hbrBack = ::CreateSolidBrush(BackColor);
    RECT rect;
    rect.left = 0;
    rect.right = Width;
    rect.top = 0;
    rect.bottom = Height;
    
    ::FillRect(hDC,&rect,hbrBack);
    
    ::DeleteObject(hbrBack);
}

void Window::Paint(HDC hDC) throw()
{
}

INT_PTR __stdcall Window::HandleWindowProcess(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam) throw()
{
    if (message == WM_NCCREATE)
    {
        LPCREATESTRUCTW lpCreate = reinterpret_cast<LPCREATESTRUCTW>(lParam);
        LPMDICREATESTRUCTW lpMdiCreate = reinterpret_cast<LPMDICREATESTRUCTW>(lpCreate->lpCreateParams);
        Window * pWindow = reinterpret_cast<Window *>(lpMdiCreate->lParam);
        pWindow->Handle = hWnd;
        pWindow->PositionX = lpCreate->x;
        pWindow->PositionY = lpCreate->y;
        pWindow->Width = lpCreate->cx;
        pWindow->Height = lpCreate->cy;
        
        ::SetPropW(hWnd,L"WindowObject",reinterpret_cast<HANDLE>(pWindow));
        
        return pWindow->WindowProcess(message,wParam,lParam);
    }
    else if (message == WM_ERASEBKGND)
    {
        Window * pObject = reinterpret_cast<Window *>(::GetPropW(hWnd,L"WindowObject"));
        if (pObject != NULL)
        {
            return 0;
        }
        else
        {
            return ::DefWindowProc(hWnd,message,wParam,lParam);
        }
    }
    else if (message == WM_PAINT)
    {
        Window * pWindow = reinterpret_cast<Window *>(::GetPropW(hWnd,L"WindowObject"));
        if (pWindow != NULL)
        {
            BITMAPINFO theInfo;
            ::memset(&theInfo,0,sizeof(BITMAPINFO));
            
            theInfo.bmiHeader.biBitCount = 24;
            theInfo.bmiHeader.biPlanes = 1;
            theInfo.bmiHeader.biWidth = pWindow->Width;
            theInfo.bmiHeader.biHeight = pWindow->Height;
            theInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
            theInfo.bmiHeader.biSizeImage = pWindow->Width * pWindow->Height * 4;
            theInfo.bmiHeader.biXPelsPerMeter = 10000;
            theInfo.bmiHeader.biYPelsPerMeter = 10000;
            
            void * lpBits = NULL;
            
            PAINTSTRUCT ps;
            HDC hPaintDC = ::BeginPaint(pWindow->Handle,&ps);
            HDC hMemDC = ::CreateCompatibleDC(hPaintDC);
            
            HBITMAP hDIBSection = ::CreateDIBSection(hPaintDC,&theInfo,DIB_RGB_COLORS,&lpBits,NULL,0);
            
            HBITMAP hOldBitmap = reinterpret_cast<HBITMAP>(::SelectObject(hMemDC,hDIBSection));
            const COLORREF clrOldText = ::SetTextColor(hMemDC,pWindow->TextColor);
            
            pWindow->EraseBackground(hMemDC);
            pWindow->Paint(hMemDC);
            
            ::StretchDIBits(hPaintDC,0,0,pWindow->Width,pWindow->Height,0,0,pWindow->Width,pWindow->Height,lpBits,&theInfo,DIB_RGB_COLORS,SRCCOPY);
            
            ::SetTextColor(hMemDC,clrOldText);
            ::SelectObject(hMemDC,hOldBitmap);
            
            ::DeleteObject(hDIBSection);
            ::DeleteDC(hMemDC);
            
            ::EndPaint(pWindow->Handle,&ps);
            
            return 0;
        }
        else
        {
            return ::DefWindowProc(hWnd,message,wParam,lParam);
        }
    }
    else
    {
        Window * pWindow = reinterpret_cast<Window *>(::GetPropW(hWnd,L"WindowObject"));
        if (pWindow != NULL)
        {
            return pWindow->WindowProcess(message,wParam,lParam);
        }
        else
        {
            return ::DefWindowProc(hWnd,message,wParam,lParam);
        }
    }
}
