#ifndef __WINDOW_H__
#define __WINDOW_H__

#ifndef __cplusplus
#error Window requires C++ compilation.
#endif // !defined __cplusplus

#include <string>
#include <vector>
#include <set>
#include <map>
#include <exception>

#include <pshpack8.h>

class Window
{
public:
    // Typedef's
    typedef std::map<unsigned long,Window *> ControlCollection;
    typedef std::wstring String;
    // Construction/Destruction
    Window() throw();
    virtual ~Window() throw();
    // Member Functions
    bool Create(const String & strName,const String & strTitle,
        unsigned long dwStyle,unsigned long dwExtendedStyle,
        long x,long y,long nWidth,long nHeight,unsigned long dwID,Window * pParent) throw();
    bool Show(int nShow) throw();
    bool Update() throw();
    bool Invalidate(bool bErase) throw();
    bool Redraw(bool bRedrawChildren) throw();
    Window * GetParent() const throw();
    Window * GetChildWindow(unsigned long dwID) const throw();
    HWND GetHwnd() const throw();
    void SetTextColor(COLORREF clrText) throw();
    COLORREF GetTextColor() const throw();
    void SetBackColor(COLORREF clrBack) throw();
    COLORREF GetBackColor() const throw();
    String GetText() const throw();
    void SetText(const String & str) throw();
    // Static Member Functions
    static bool RegisterWindowClass(const String & strName,HCURSOR hCursor,HICON hIcon,HICON hIconSm) throw();
    static bool UnregisterWindowClass(const String & strName) throw();
protected:
    // Member Variables
    HWND Handle;
    HMENU MenuHandle;
    HFONT TextFont;
    long PositionX;
    long PositionY;
    long Width;
    long Height;
    COLORREF TextColor;
    COLORREF BackColor;
    std::wstring Text;
    ControlCollection Controls;
    // Member Functions
    virtual INT_PTR WindowProcess(UINT message,WPARAM wParam,LPARAM lParam) throw();
    virtual void EraseBackground(HDC hPaintDC) throw();
    virtual void Paint(HDC hPaintDC) throw();
private:
    // Static Member Functions
    static INT_PTR __stdcall HandleWindowProcess(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam) throw();
};

#include <poppack.h>

#endif // !defined __WINDOW_H__
